/// @file
////////////////////////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2019/20      Christian Lessig, Otto-von-Guericke Universitaet Magdeburg
///
////////////////////////////////////////////////////////////////////////////////////////////////////
///
///  module     : lecture 5
///
///  author     : lessig@isg.cs.ovgu.de
///
///  project    : GPU Programming
///
///  description: race conditions
///
////////////////////////////////////////////////////////////////////////////////////////////////////

// includes, system
#include <iostream>
#include <vector>
#include <thread>
#include <mutex>
#include <condition_variable>
#include <chrono>
#include <ctime>


bool stop = false;


struct Buffer {

  Buffer( int size) :
    occupied( 0),
    nextin( 0),
    nextout( 0),
    data( size)
  {}

  std::mutex mutex;
  std::condition_variable  producer;
  std::condition_variable  consumer;

  std::size_t occupied;
  std::size_t nextin;
  std::size_t nextout;

  // data
  std::vector<int> data;
};

void
producer( Buffer& buf) {

  while( ! stop ) {

    // generate next item: expensive process so run asynchronously
    int next_item = rand();

    std::unique_lock<std::mutex>  lock( buf.mutex);

    if( buf.occupied >= buf.data.size()) {
      buf.producer.wait( lock, [&]{ return buf.occupied < buf.data.size(); } );
    }

    buf.data[buf.nextin] = next_item;

    buf.nextin++;
    buf.nextin %= buf.data.size();
    buf.occupied++;

    buf.consumer.notify_one();
  }

}

void
consumer( Buffer& buf) {

  while( ! stop ) {

    int next_item = 0;

    // access next item from shared buffer

    {
      std::unique_lock<std::mutex>  lock( buf.mutex);

      // check if buffer is empty, if necessary wait
      if( 0 == buf.occupied) {
        buf.consumer.wait( lock, [&]{ return buf.occupied > 0; });
      }

      // get element
      next_item = buf.data[buf.nextout];

      // update buffer
      buf.nextout++;
      buf.nextout %= buf.data.size();

      buf.occupied--;
      buf.producer.notify_one();
    }

    // process item: expensive so run asynchronously
    next_item++;
  }

}

int
main( int /*argc*/, char** /*argv*/) {

  Buffer buf( 10);
  int num_consumers = 2;
  int num_producers = 2;

  std::vector<std::thread> consumers;
  for( int it = 0; it < num_consumers; ++it) {
    consumers.push_back( std::thread( consumer, std::ref( buf)));
    std::cout << "Created consumer " << it << std::endl;
  }

  std::vector<std::thread> producers;
  for( int it = 0; it < num_producers; ++it) {
    consumers.push_back( std::thread( producer, std::ref( buf)));
    std::cout << "Created producer " << it << std::endl;
  }

  // wait and let them work
  using std::chrono::system_clock;
  std::time_t tt = system_clock::to_time_t (system_clock::now());
  struct std::tm * ptm = std::localtime(&tt);
  ptm->tm_min++;
  ptm->tm_sec=0;
  std::this_thread::sleep_until (system_clock::from_time_t (mktime(ptm)));
  // tell producers and consumers to stop
  stop = true;

  for( auto& th : consumers) {
    th.join();
  }
  for( auto& th : producers) {
    th.join();
  }

  return 0;
}
